Dim QMetryURL			' QMetry URL. Example: http://login.qmetry.com/qmetry
Dim QMetryInstance		' QMetry Instance Name. 
						' if you are accessing QMetry like http://login.qmetry.com/qmetry4.0, then instance name will be qmetry4.0
Dim objHTTP				' Object for performing HTTP Requests
Dim XMLDoc				' Hold Schedule XML DOM Structure
Dim XMLDocApi			' XML Doc object for parsing WS API Response

Dim ScheduleXMLFile		' Variable to hold value of Schedule XML Files provided by QMetry Agent
Dim FileSystemObj		' File System Object to get information about Schedule XML File

Dim QMetryUsername		' QMetry Username, used to update status of Test Execution
Dim QMetryPassword		' QMetry Password, used to update status of Test Execution
Dim ConfigLocation		' Holds value of location of QMetry Instance config file
Dim QMetryUpdateStepResult	' If 1, then QTP launcher will update Test Case Steps in QMetry

Dim StepArray			' Holds all Test Case Steps
Dim StepStatusArray		' Holds status of all Test Case Steps from QTP Result file
Dim StepResultArray		' Holds Actual Result/Description of Test Case Step from QTP Result file

Dim TestDataVariables	' Holds Test Data Variables returned by getTestDataVariables API
Dim TestDataVar			' Holds single Test Data Object
Dim TestDataVarDict		' Dictionary object which holds all Test Data for a Test Case

Dim HTTPProxyEnabled	' Flag which holds value of whether HTTP Proxy is enabled or not (true = enabled)
Dim HTTPProxyHost		' Hostname/IP Address of HTTP Proxy Server
Dim HTTPProxyPort		' HTTP Proxy Server Port
Dim HTTPProxyUserName 	' HTTP Proxy Server Username if it requires authentication
Dim HTTPProxyPassword	' HTTP Proxy Server Password

' Read QMetry configuration from config file
Set FileSystemObj = CreateObject("Scripting.FileSystemObject")
ConfigLocation = FileSystemObj.GetFile(Wscript.ScriptFullName).ParentFolder
Dim fso, json, str, o, i
Set json = New VbsJson
Set fso = WScript.CreateObject("Scripting.Filesystemobject")

If FileSystemObj.FileExists(ConfigLocation & "\QMetryConfig.properties") Then
	LoadQMetryConfig (ConfigLocation & "\QMetryConfig.properties")
	If QMetryURL = "" Then
		WScript.StdOut.Write vbNewline & "ERROR: QMetryHost not Set in " & ConfigLocation & "\QMetryConfig.properties"
		WScript.Quit 1
	End If
	If QMetryUsername = "" Then
		WScript.StdOut.Write vbNewline & "ERROR: QMetryUsername not Set in " & ConfigLocation & "\QMetryConfig.properties"
		WScript.Quit 1
	End If
	If QMetryPassword = "" Then
		WScript.StdOut.Write vbNewline & "ERROR: QMetryPassword not Set in " & ConfigLocation & "\QMetryConfig.properties"
		WScript.Quit 1
	End If	
Else
	WScript.StdOut.Write vbNewline & "ERROR: Can not find " & ConfigLocation & "\QMetryConfig.properties"
	WScript.Quit 1
End If


' Check Schedule XML file
If WScript.Arguments.Count > 0 Then
	ScheduleXMLFile = Wscript.Arguments(0)
	If ScheduleXMLFile = "" Then
		WScript.StdOut.Write vbNewline & "Schedule XML File not provided, nothing to parse"
		WScript.Quit 1
	Else
		If FileSystemObj.FileExists(ScheduleXMLFile) Then
			WScript.StdOut.Write vbNewline & "Parsing Schedule XML: " & ScheduleXMLFile
		Else
			WScript.StdOut.Write vbNewline & "Schedule XML File: " & ScheduleXMLFile & " does not exists"
			WScript.Quit 1
		End If
	End If
Else
	WScript.StdOut.Write vbNewline & "Schedule XML File not provided, nothing to parse"
	WScript.Quit 1
End If


str = fso.OpenTextFile(ScheduleXMLFile).ReadAll
Set o = json.Decode(str)

ScheduleId = o("schedule")("schdeuleId")
ProjectName = o("schedule")("projectName")
ReleaseName = o("schedule")("releaseName")
BuildName = o("schedule")("buildName")
Project = o("schedule")("projectId")
Release = o("schedule")("releaseId")
Build = o("schedule")("buildId")
TestSuiteId = o("schedule")("testsuiteId")
PlatformId = o("schedule")("platformId")

TestSuiteRunId = o("schedule")("testsuiteRunId")
DropId = o("schedule")("dropId")

WScript.StdOut.Write vbNewline & "Schedule Id: " & ScheduleId
WScript.StdOut.Write vbNewline & "Project: " & ProjectName
WScript.StdOut.Write vbNewline & "Release: " & ReleaseName
WScript.StdOut.Write vbNewline & "Build: " & BuildName
WScript.StdOut.Write vbNewline & "TestSuiteId: " & TestSuiteId
WScript.StdOut.Write vbNewline & "Platform ID: " & PlatformId

Set objHTTP = CreateObject("MSXML2.ServerXMLHTTP.6.0")

' Fetch Test Case information from Schedule XML File
AllTestCases = o("schedule")("testcases")
TotalTestCases = 0

For Each tc In o("schedule")("testcases")
	TotalTestCases = TotalTestCases + 1
Next

WScript.StdOut.Write vbNewline & "TestCases to Execute: " & TotalTestCases

' Launch QTP with appropriate Test
Set QTPApp = CreateObject("QuickTest.Application")
If QTPApp.Launched <> True then 'Launch QTP application
	QTPApp.Launch
	If Err.Number <> 0 Then
		Set QTPApp = Nothing
		WScript.StdOut.Write vbNewline & "Error occured while Attempting to launch QTP" & vbNewLine & Err.Description
		WScript.StdOut.Write vbNewline & "QMetry Agent can not execute any QTP Test Script, exiting" & vbNewLine & Err.Description
		WScript.Quit 1
	End If
End If

' Login to QMetry using WS API and get authentication token
WScript.StdOut.Write vbNewline & vbNewline & "Authenticating with QMetry Application..." & vbNewline
Token = Login (QMetryUsername, QMetryPassword)
'Token = "AbPU4SgihnR3oECNgBwZ"
WScript.StdOut.Write vbNewline & "Token: " & Token
If IsNull(Token) Then
	WScript.StdOut.Write vbNewline & vbNewline & "Can not Login to QMetry, Exiting..."
	WScript.Quit 1
End If
WScript.StdOut.Write vbNewline & vbNewline

' Update Test Suite Start Time
If Token <> "ERROR" Then
	Msg = UpdateTestSuiteTime (Token, "start", TestSuiteRunId, PlatformId)
End If

' Execute Test Cases one by one
For Each tc In o("schedule")("testcases")
	
	Set TestCase = tc
	TestCaseRunId = TestCase("tcrunId")
	TestCaseId = TestCase("testcaseId")
	WScript.StdOut.Write vbNewline & "	Test Case ID: " & TestCaseId & ", Run ID: " & TestCaseRunId & vbNewline

	TestCaseScript = TestCase("testScriptName")
	
	WScript.StdOut.Write vbNewline & "	Test Case Script: " & TestCaseScript
	
	If TestCaseScript = "" Then
		WScript.StdOut.Write vbNewline & "	No associated Test Script with Test Case: " & TestCaseId & ", Skipping this Test Case Execution" & vbNewline
	ElseIf Not FileSystemObj.FolderExists(TestCaseScript) Then
		WScript.StdOut.Write vbNewline & "	Test Script " & TestCaseScript & " does not exists for Test Case Id " & TestCaseId & ", Skipping this Test Case Execution" & vbNewline
	Else
		' Open QTP Test in read only mode
		QTPApp.Open TestCaseScript, True
		QTPApp.Visible=False
		QTPApp.Options.Run.RunMode = "Fast"
		QTPApp.Options.Run.ViewResults = False

		' Set run settings for the test
		Set qtTest = QTPApp.Test
		
		qtTest.Settings.Run.OnError = "NextStep" ' Instruct QuickTest to perform next step when error occurs

		Set qtResultsOpt = CreateObject("QuickTest.RunResultsOptions") ' Create the Run Results Options object
		QtpResultLocation = ConfigLocation & "\QMetryQTP_TC" & TestCaseId
		qtResultsOpt.ResultsLocation = QtpResultLocation ' Set the results location
		
		qtTest.Run qtResultsOpt, true ' Run the test

		TestCaseStatus = qtTest.LastRunResults.Status
'		TestCaseStatus = "passed"

		
		If TestCaseStatus = "Failed" Then
			TestCaseStatus = "failed"
		Else
			TestCaseStatus = "passed"
		End If
		
		WScript.StdOut.Write vbNewline & "Test Case Result: " & TestCaseStatus ' Check the results of the test run

		Result = ExecuteTestCaseUsingRunId (Token, TestCaseRunId, o("schedule")("runstats")(TestCaseStatus), TestSuiteRunId, DropId)
		
		Set qtResultsOpt = Nothing ' Release the Run Results Options object
		qtTest.Close 	' Close the test
		Set qtTest = Nothing ' Release the Test object
		
	End If
Next

' Update Test Suite End Time
If Token <> "ERROR" Then
	'Msg = UpdateTestSuiteTime (Token, "End", TestSuiteRunId, PlatformId)
End If

WScript.StdOut.Write vbNewline & "Closing QTP..." 

QTPApp.quit		' Quit from QTP
Set QTPApp = Nothing

WScript.StdOut.Write "Closed." 

'''
''' QMetry Web Services Related Function
'''
' Login to QMetry Application and return Authentication Token
Function jsonDecode(str)
	Dim jsonparser
	Set jsonparser = New VbsJson
	jsonDecode = jsonparser.Decode(str)
End Function

Function Login (username, password)
	'Create the SOAP Envelope
	Envelope = "username=" & username & "&password=" & password

	'Set a header for the method to be called
	objHTTP.open "POST", QMetryURL & "/rest/login", False

	'Set HTTP Request header and Proxy settings
	objHTTP.setRequestHeader "Accept", "application/json; charset=UTF-8"
	objHTTP.setRequestHeader "Content-Type", "application/x-www-form-urlencoded; charset=UTF-8"

	'Make the SOAP call
	objHTTP.send Envelope

	'Get the return envelope
	strReturn = objHTTP.responseText
	WScript.Echo strReturn
	Set loginRes = json.Decode(strReturn)
	AuthToken = loginRes("usertoken")
	Login = AuthToken
End Function

' Set the scope of your testing (i.e. which Project, Release and Build)
Function SetScope ()
	'Set a header for the method to be called
	objHTTP.setRequestHeader "scope", Project & ":" & Release & ":" & Build
	objHTTP.setRequestHeader "usertoken", Token

End Function

' Execute the Test Case 
Function ExecuteTestCaseUsingRunId (Token, TestCaseRunId, ExecStatus, TestSuiteRunId, DropId)
	'Create the SOAP Envelope
'	Token = "AbPU4SgihnR3oECNgBwZ"
	
	WScript.StdOut.Write vbNewline & "Updating the result with Token :"& Token & vbNewline
	
	If ExecStatus = "Failed" Then
		TestCaseStatus = "failed"
	Else
		TestCaseStatus = "passed"
	End If

	If IsNull(DropId) or DropId="" Then
		DropId = chr(34) & DropId & chr(34)
	End If
		
	Envelope = "{""dropID"":"& DropId &",""entityIDs"":"& chr(34) & TestCaseRunId & chr(34) &",""qmRunObj"":"""",""qmTsRunId"":"& chr(34) & TestSuiteRunId & chr(34) &",""entityType"":""TCR"",""runStatusID"":"& ExecStatus &"}"

	WScript.StdOut.Write vbNewline & "Envelope :" & Envelope & vbNewline

	
	'Set a header for the method to be called
	objHTTP.open "PUT", QMetryURL & "/rest/execution/runstatus/bulkupdate", False

	'Set HTTP Request header and Proxy settings
	SetHTTPRequestHeader(objHTTP)
	SetScope()
	objHTTP.setRequestHeader "Referer", QMetryURL
	objHTTP.setRequestHeader "usertoken", Token

	'WScript.StdOut.Write vbNewline & "Header :" & objHTTP.getRequestHeader("usertoken")& vbNewline
	

	'Make the SOAP call
	objHTTP.send Envelope

	'Get the return envelope
	strReturn = objHTTP.responseText

	WScript.StdOut.Write vbNewline & "Updating the result response : "& strReturn & vbNewline

	Envelope = "{""entityIDs"":"& TestCaseRunId &",""qmRunObj"":"""",""field"":""comment"",""entityType"":""TCR"",""fieldVal"":""Updated By QMetry Automation Agent""}"

	objHTTP.open "PUT", QMetryURL & "/rest/execution/updaterun", False
	SetScope()
	SetHTTPRequestHeader(objHTTP)
	objHTTP.send Envelope
	
	WScript.StdOut.Write vbNewline & "Updating the run : "& objHTTP.responseText & vbNewline
	
	ExecuteTestCaseUsingRunId = strReturn
End Function

' Update Test Suite Start/EndTime
Function UpdateTestSuiteTime (Token, startOrEnd, TestSuiteRunId, PlatformId)
	MethodName = "start"
	If startOrEnd = "End" Then
		MethodName = "end"
	End If
		
	'Create the SOAP Envelope
	Envelope = "{""entityType"":""tsrun"",""entityId"":"& TestSuiteRunId &",""timeType"":"&MethodName&"}"

	'Set a header for the method to be called
	objHTTP.open "PUT", QMetryURL & "/rest/execution/updateexecutiontime", False

	'Set HTTP Request header and Proxy settings
	SetHTTPRequestHeader(objHTTP)

	'Make the SOAP call
	objHTTP.send Envelope

	'Get the return envelope
	strReturn = objHTTP.responseText
	UpdateTestSuiteTime = strReturn
End Function

' Update the Test Case execution Status
Function UpdateSchedulerStatus (Token, ScheduleId, Status)
	'Create the SOAP Envelope
	Envelope = "<soapenv:Envelope xmlns:soapenv=""http://schemas.xmlsoap.org/soap/envelope/"" xmlns:xsd=""http://www.wso2.org/php/xsd""><soapenv:Header/><soapenv:Body><xsd:updateSchedulerComments><xsd:token>" & Token & "</xsd:token><xsd:scheduleId>" & ScheduleId & "</xsd:scheduleId><xsd:comments>" & Status & "</xsd:comments></xsd:updateSchedulerComments></soapenv:Body></soapenv:Envelope>"

	'Set HTTP Request header and Proxy settings
	SetHTTPRequestHeader(objHTTP)

	'Set a header for the method to be called
	objHTTP.setRequestHeader "SOAPAction", "http://" & QMetryURL & "/WEB-INF/ws/service.php/updateSchedulerComments"

	'Make the SOAP call
	objHTTP.send Envelope

	'Get the return envelope
	strReturn = objHTTP.responseText
	SOAPXML = ExtractEnvelope (strReturn)
	'WScript.StdOut.Write SOAPXML

	'Load the return envelope into a DOM
	XMLDoc.loadXML (SOAPXML)

	Set FaultNode = XMLDoc.getElementsByTagName("faultstring")
	If FaultNode.length = 0 Then
		'Query the return envelope
		Set XMLNode = XMLDoc.getElementsByTagName("ns1:success")
		For Each Elem In XMLNode
			Message = Elem.firstChild.nodeValue
		Next
		'WScript.StdOut.Write vbNewline & MethodName & ": " & Message
	Else
		For Each Elem In FaultNode
			FaultString = Elem.firstChild.nodeValue
		Next
		WScript.StdOut.Write vbNewline & "ERROR: " & FaultString
		Message = "ERROR"
	End If
	UpdateSchedulerStatus = Message
End Function

' Load QMetry Configuration file and fetch parameters
Sub LoadQMetryConfig (FileName)
 DIM strConfigLine
 DIM fConFile
 DIM EqPos, strLen, varName, varVal
 SET fConFile = FileSystemObj.OpenTextFile(FileName)
 WHILE NOT fConFile.AtEndOfStream
 	strConfigLine = fConFile.ReadLine
	strConfigLine = TRIM(strConfigLine)
	IF (INSTR(1,strConfigLine,"#",1) <> 1 AND LEN(strConfigLine) <> 0) THEN
		EqPos = INSTR(1,strConfigLine,"=",1)
		strLen = LEN(strConfigLine)
		varName = LCASE(TRIM(MID(strConfigLine, 1, EqPos - 1)))
		varVal = TRIM(MID(strConfigLine, EqPos + 1, strLen - EqPos))
		SELECT CASE varName
			'ADD EACH OCCURRENCE OF THE CONFIGURATION FILE VARIABLES(KEYS)
			CASE LCASE("qmetryurl")
				IF varVal <> "" THEN QMetryURL = varVal
			CASE LCASE("qmetryusername")
				IF varVal <> "" THEN QMetryUsername = varVal
			CASE LCASE("qmetrypassword")
				IF varVal <> "" THEN QMetryPassword = varVal
			CASE LCASE("updatestepresult")
				IF varVal <> "" THEN QMetryUpdateStepResult = varVal
			CASE LCASE("setproxy")
				IF varVal <> "" THEN HTTPProxyEnabled = LCASE(varVal)
			CASE LCASE("proxyhost")
				IF varVal <> "" THEN HTTPProxyHost = varVal
			CASE LCASE("proxyport")
				IF varVal <> "" THEN HTTPProxyPort = varVal
			CASE LCASE("proxyuser")
				IF varVal <> "" THEN HTTPProxyUserName = varVal
			CASE LCASE("proxypassword")
				IF varVal <> "" THEN HTTPProxyPassword = varVal				
		END SELECT
	END IF
 WEND
 fConFile.Close
End Sub

Function Base64Encode(sText)
    Dim oXML, oNode

    Set oXML = CreateObject("MSXML2.DOMDocument.3.0")
    Set oNode = oXML.CreateElement("base64")
    oNode.dataType = "bin.base64"
    oNode.nodeTypedValue = Stream_StringToBinary(sText)
    
	Base64Encode = oNode.text
	
    Set oNode = Nothing
    Set oXML = Nothing
End Function

Function Stream_StringToBinary(Text)
  Const adTypeText = 2
  Const adTypeBinary = 1

  'Create Stream object
  Dim BinaryStream 'As New Stream
  Set BinaryStream = CreateObject("ADODB.Stream")

  'Specify stream type - we want To save text/string data.
  BinaryStream.Type = adTypeText

  'Specify charset For the source text (unicode) data.
  BinaryStream.CharSet = "us-ascii"

  'Open the stream And write text/string data To the object
  BinaryStream.Open
  BinaryStream.WriteText Text

  'Change stream type To binary
  BinaryStream.Position = 0
  BinaryStream.Type = adTypeBinary

  'Ignore first two bytes - sign of
  BinaryStream.Position = 0

  'Open the stream And get binary data from the object
  Stream_StringToBinary = BinaryStream.Read

  Set BinaryStream = Nothing
End Function

Function SetHTTPRequestHeader (objHTTP)
	objHTTP.setRequestHeader "Content-Type", "application/json; charset=UTF-8"
	objHTTP.setRequestHeader "Accept", "application/json; charset=UTF-8"
End Function

Function ExtractEnvelope (envelope)
	BodyStartPos = InStr (envelope, "<soapenv:Envelope ")
	BodyEndPos = InStr (envelope, "</soapenv:Envelope>")+19

	If BodyStartPos = 0 Then
		WScript.StdOut.Write vbNewline & "ERROR: Cannot parse Envelope (Envelope start tag not found)"
		WScript.StdOut.Write vbNewline & envelope
		WScript.StdOut.Write vbNewline
		SOAPXML = ""
	ElseIf BodyEndPos = 19 Then
		WScript.StdOut.Write vbNewline & "ERROR: Cannot parse Envelope (Envelope end tag not found)"
		WScript.StdOut.Write vbNewline & envelope
		WScript.StdOut.Write vbNewline
		SOAPXML = ""
	Else
		SOAPXML = Mid (envelope, BodyStartPos, BodyEndPos-BodyStartPos)
	End If
	
	ExtractEnvelope = SOAPXML
End Function

Class VbsJson
	'Author: Demon
	'Date: 2012/5/3
	'Website: http://demon.tw
	Private Whitespace, NumberRegex, StringChunk
	Private b, f, r, n, t

	Private Sub Class_Initialize
		Whitespace = " " & vbTab & vbCr & vbLf
		b = ChrW(8)
		f = vbFormFeed
		r = vbCr
		n = vbLf
		t = vbTab

		Set NumberRegex = New RegExp
		NumberRegex.Pattern = "(-?(?:0|[1-9]\d*))(\.\d+)?([eE][-+]?\d+)?"
		NumberRegex.Global = False
		NumberRegex.MultiLine = True
		NumberRegex.IgnoreCase = True

		Set StringChunk = New RegExp
		StringChunk.Pattern = "([\s\S]*?)([""\\\x00-\x1f])"
		StringChunk.Global = False
		StringChunk.MultiLine = True
		StringChunk.IgnoreCase = True
	End Sub
	
	'Return a JSON string representation of a VBScript data structure
	'Supports the following objects and types
	'+-------------------+---------------+
	'| VBScript          | JSON          |
	'+===================+===============+
	'| Dictionary        | object        |
	'+-------------------+---------------+
	'| Array             | array         |
	'+-------------------+---------------+
	'| String            | string        |
	'+-------------------+---------------+
	'| Number            | number        |
	'+-------------------+---------------+
	'| True              | true          |
	'+-------------------+---------------+
	'| False             | false         |
	'+-------------------+---------------+
	'| Null              | null          |
	'+-------------------+---------------+
	Public Function Encode(ByRef obj)
		Dim buf, i, c, g
		Set buf = CreateObject("Scripting.Dictionary")
		Select Case VarType(obj)
			Case vbNull
				buf.Add buf.Count, "null"
			Case vbBoolean
				If obj Then
					buf.Add buf.Count, "true"
				Else
					buf.Add buf.Count, "false"
				End If
			Case vbInteger, vbLong, vbSingle, vbDouble
				buf.Add buf.Count, obj
			Case vbString
				buf.Add buf.Count, """"
				For i = 1 To Len(obj)
					c = Mid(obj, i, 1)
					Select Case c
						Case """" buf.Add buf.Count, "\"""
						Case "\"  buf.Add buf.Count, "\\"
						Case "/"  buf.Add buf.Count, "/"
						Case b    buf.Add buf.Count, "\b"
						Case f    buf.Add buf.Count, "\f"
						Case r    buf.Add buf.Count, "\r"
						Case n    buf.Add buf.Count, "\n"
						Case t    buf.Add buf.Count, "\t"
						Case Else
							If AscW(c) >= 0 And AscW(c) <= 31 Then
								c = Right("0" & Hex(AscW(c)), 2)
								buf.Add buf.Count, "\u00" & c
							Else
								buf.Add buf.Count, c
							End If
					End Select
				Next
				buf.Add buf.Count, """"
			Case vbArray + vbVariant
				g = True
				buf.Add buf.Count, "["
				For Each i In obj
					If g Then g = False Else buf.Add buf.Count, ","
					buf.Add buf.Count, Encode(i)
				Next
				buf.Add buf.Count, "]"
			Case vbObject
				If TypeName(obj) = "Dictionary" Then
					g = True
					buf.Add buf.Count, "{"
					For Each i In obj
						If g Then g = False Else buf.Add buf.Count, ","
						buf.Add buf.Count, """" & i & """" & ":" & Encode(obj(i))
					Next
					buf.Add buf.Count, "}"
				Else
					Err.Raise 8732,,"None dictionary object"
				End If
			Case Else
				buf.Add buf.Count, """" & CStr(obj) & """"
		End Select
		Encode = Join(buf.Items, "")
	End Function

	'Return the VBScript representation of ``str(``
	'Performs the following translations in decoding
	'+---------------+-------------------+
	'| JSON          | VBScript          |
	'+===============+===================+
	'| object        | Dictionary        |
	'+---------------+-------------------+
	'| array         | Array             |
	'+---------------+-------------------+
	'| string        | String            |
	'+---------------+-------------------+
	'| number        | Double            |
	'+---------------+-------------------+
	'| true          | True              |
	'+---------------+-------------------+
	'| false         | False             |
	'+---------------+-------------------+
	'| null          | Null              |
	'+---------------+-------------------+
	Public Function Decode(ByRef str)
		Dim idx
		idx = SkipWhitespace(str, 1)

		If Mid(str, idx, 1) = "{" Then
			Set Decode = ScanOnce(str, 1)
		Else
			Decode = ScanOnce(str, 1)
		End If
	End Function
	
	Private Function ScanOnce(ByRef str, ByRef idx)
		Dim c, ms

		idx = SkipWhitespace(str, idx)
		c = Mid(str, idx, 1)

		If c = "{" Then
			idx = idx + 1
			Set ScanOnce = ParseObject(str, idx)
			Exit Function
		ElseIf c = "[" Then
			idx = idx + 1
			ScanOnce = ParseArray(str, idx)
			Exit Function
		ElseIf c = """" Then
			idx = idx + 1
			ScanOnce = ParseString(str, idx)
			Exit Function
		ElseIf c = "n" And StrComp("null", Mid(str, idx, 4)) = 0 Then
			idx = idx + 4
			ScanOnce = Null
			Exit Function
		ElseIf c = "t" And StrComp("true", Mid(str, idx, 4)) = 0 Then
			idx = idx + 4
			ScanOnce = True
			Exit Function
		ElseIf c = "f" And StrComp("false", Mid(str, idx, 5)) = 0 Then
			idx = idx + 5
			ScanOnce = False
			Exit Function
		End If
		
		Set ms = NumberRegex.Execute(Mid(str, idx))
		If ms.Count = 1 Then
			idx = idx + ms(0).Length
			ScanOnce = CDbl(ms(0))
			Exit Function
		End If
		
		Err.Raise 8732,,"No JSON object could be ScanOnced"
	End Function

	Private Function ParseObject(ByRef str, ByRef idx)
		Dim c, key, value
		Set ParseObject = CreateObject("Scripting.Dictionary")
		idx = SkipWhitespace(str, idx)
		c = Mid(str, idx, 1)
		
		If c = "}" Then
			Exit Function
		ElseIf c <> """" Then
			Err.Raise 8732,,"Expecting property name"
		End If

		idx = idx + 1
		
		Do
			key = ParseString(str, idx)

			idx = SkipWhitespace(str, idx)
			If Mid(str, idx, 1) <> ":" Then
				Err.Raise 8732,,"Expecting : delimiter"
			End If

			idx = SkipWhitespace(str, idx + 1)
			If Mid(str, idx, 1) = "{" Then
				Set value = ScanOnce(str, idx)
			Else
				value = ScanOnce(str, idx)
			End If
			ParseObject.Add key, value

			idx = SkipWhitespace(str, idx)
			c = Mid(str, idx, 1)
			If c = "}" Then
				Exit Do
			ElseIf c <> "," Then
				Err.Raise 8732,,"Expecting , delimiter"
			End If

			idx = SkipWhitespace(str, idx + 1)
			c = Mid(str, idx, 1)
			If c <> """" Then
				Err.Raise 8732,,"Expecting property name"
			End If

			idx = idx + 1
		Loop

		idx = idx + 1
	End Function
	
	Private Function ParseArray(ByRef str, ByRef idx)
		Dim c, values, value
		Set values = CreateObject("Scripting.Dictionary")
		idx = SkipWhitespace(str, idx)
		c = Mid(str, idx, 1)

		If c = "]" Then
			ParseArray = values.Items
			Exit Function
		End If

		Do
			idx = SkipWhitespace(str, idx)
			If Mid(str, idx, 1) = "{" Then
				Set value = ScanOnce(str, idx)
			Else
				value = ScanOnce(str, idx)
			End If
			values.Add values.Count, value

			idx = SkipWhitespace(str, idx)
			c = Mid(str, idx, 1)
			If c = "]" Then
				Exit Do
			ElseIf c <> "," Then
				Err.Raise 8732,,"Expecting , delimiter"
			End If

			idx = idx + 1
		Loop

		idx = idx + 1
		ParseArray = values.Items
	End Function
	
	Private Function ParseString(ByRef str, ByRef idx)
		Dim chunks, content, terminator, ms, esc, char
		Set chunks = CreateObject("Scripting.Dictionary")

		Do
			Set ms = StringChunk.Execute(Mid(str, idx))
			If ms.Count = 0 Then
				Err.Raise 8732,,"Unterminated string starting"
			End If
			
			content = ms(0).Submatches(0)
			terminator = ms(0).Submatches(1)
			If Len(content) > 0 Then
				chunks.Add chunks.Count, content
			End If
			
			idx = idx + ms(0).Length
			
			If terminator = """" Then
				Exit Do
			ElseIf terminator <> "\" Then
				Err.Raise 8732,,"Invalid control character"
			End If
			
			esc = Mid(str, idx, 1)

			If esc <> "u" Then
				Select Case esc
					Case """" char = """"
					Case "\"  char = "\"
					Case "/"  char = "/"
					Case "b"  char = b
					Case "f"  char = f
					Case "n"  char = n
					Case "r"  char = r
					Case "t"  char = t
					Case Else Err.Raise 8732,,"Invalid escape"
				End Select
				idx = idx + 1
			Else
				char = ChrW("&H" & Mid(str, idx + 1, 4))
				idx = idx + 5
			End If

			chunks.Add chunks.Count, char
		Loop

		ParseString = Join(chunks.Items, "")
	End Function

	Private Function SkipWhitespace(ByRef str, ByVal idx)
		Do While idx <= Len(str) And _
			InStr(Whitespace, Mid(str, idx, 1)) > 0
			idx = idx + 1
		Loop
		SkipWhitespace = idx
	End Function

End Class